from __future__ import annotations

from datetime import datetime, timezone
from os import PathLike
from typing import Annotated, TypeAlias, TypeVar, Union

from pydantic import AfterValidator

T = TypeVar("T")

StrPath: TypeAlias = Union[str, PathLike[str]]
"""String or pathlib.Path"""

UTCDateTime = Annotated[datetime, AfterValidator(lambda dt: dt.astimezone(timezone.utc))]
"""datetime that's always in UTC."""



from __future__ import annotations

import shutil
from pathlib import Path

from nzb import NZBMetaEditor, NZBParser

nzbs = Path("tests/__nzbs__").resolve()
encoding = "utf-8"


def test_meta_clear(tmp_path: Path) -> None:
    edited = nzbs / "spec_example_meta_clear.nzb"
    out = NZBMetaEditor.from_file(nzbs / "spec_example.nzb").clear().save(tmp_path / "spec_example_meta_clear.nzb")
    assert out.read_text(encoding).strip() == edited.read_text(encoding).strip()


def test_nzb_with_no_head_clear(tmp_path: Path) -> None:
    nzb = nzbs / "nzb_with_no_head.nzb"
    out = NZBMetaEditor.from_file(nzb).clear().save(tmp_path / "nzb_with_no_head.nzb")
    assert out.is_file()


def test_meta_remove_append(tmp_path: Path) -> None:
    edited = nzbs / "spec_example_meta_append.nzb"
    out = (
        NZBMetaEditor.from_file(nzbs / "spec_example.nzb")
        .remove("password")
        .append(passwords="new secret!")
        .save(tmp_path / "spec_example.nzb")
    )
    assert out.read_text(encoding).strip() == edited.read_text(encoding).strip()


def test_meta_append_when_file_has_no_meta(tmp_path: Path) -> None:
    append = NZBMetaEditor.from_file(nzbs / "no_meta.nzb").append(title="appending").save(tmp_path / "no_meta.nzb")
    set = NZBMetaEditor.from_file(nzbs / "no_meta.nzb").set(title="appending").save(tmp_path / "no_meta.nzb")
    assert append.read_text(encoding).strip() == set.read_text(encoding).strip()


def test_meta_append_when_file_has_single_meta(tmp_path: Path) -> None:
    append = (
        NZBMetaEditor.from_file(nzbs / "single_meta.nzb").append(title="appending").save(tmp_path / "single_meta.nzb")
    )
    parsed = NZBParser.from_file(append).parse()
    assert parsed.meta.title == "appending"


def test_meta_remove_empty(tmp_path: Path) -> None:
    rm = (
        NZBMetaEditor.from_file(nzbs / "spec_example_meta_clear.nzb")
        .remove("category")
        .save(tmp_path / "spec_example_meta_clear.nzb")
    )
    parsed = NZBParser.from_file(rm).parse()
    assert parsed.meta.title is None
    assert parsed.meta.passwords is None
    assert parsed.meta.tags is None
    assert parsed.meta.category is None


def test_meta_remove_one(tmp_path: Path) -> None:
    rm = NZBMetaEditor.from_file(nzbs / "single_meta.nzb").remove("title").save(tmp_path / "single_meta.nzb")
    parsed = NZBParser.from_file(rm).parse()
    assert parsed.meta.title is None
    assert parsed.meta.passwords is None
    assert parsed.meta.tags is None
    assert parsed.meta.category is None


def test_meta_remove_missing_key(tmp_path: Path) -> None:
    rm = NZBMetaEditor.from_file(nzbs / "single_meta.nzb").remove("akldakldjakldjs").save(tmp_path / "single_meta.nzb")
    parsed = NZBParser.from_file(rm).parse()
    assert parsed.meta.title is not None
    assert parsed.meta.passwords is None
    assert parsed.meta.tags is None
    assert parsed.meta.category is None


def test_meta_set(tmp_path: Path) -> None:
    edited = nzbs / "spec_example_meta_set.nzb"
    out = (
        NZBMetaEditor.from_file(nzbs / "spec_example.nzb")
        .set(title="New title", tags=["test", "test2"])
        .save(tmp_path / "spec_example.nzb")
    )
    assert out.read_text(encoding).strip() == edited.read_text(encoding).strip()


def test_meta_set_empty(tmp_path: Path) -> None:
    edited = nzbs / "spec_example_meta_set.nzb"
    out = NZBMetaEditor.from_file(nzbs / "spec_example_meta_set.nzb").set().save(tmp_path / "spec_example_meta_set.nzb")
    assert out.read_text(encoding).strip() == edited.read_text(encoding).strip()


def test_meta_save_overwrite(tmp_path: Path) -> None:
    original = nzbs / "no_meta.nzb"
    tmp_nzb: Path = shutil.copy(original, tmp_path / "no_meta.nzb")
    NZBMetaEditor.from_file(tmp_nzb).save(overwrite=True)
    assert original.read_text(encoding).strip() == tmp_nzb.read_text(encoding).strip()


def test_no_doctype(tmp_path: Path) -> None:
    original = nzbs / "no_doctype.nzb"
    tmp_nzb: Path = shutil.copy(original, tmp_path / "no_doctype.nzb")
    NZBMetaEditor.from_file(tmp_nzb).save(overwrite=True)
    assert original.read_text(encoding).strip() == tmp_nzb.read_text(encoding).strip()



from __future__ import annotations

from pathlib import Path

import pytest

from nzb import InvalidNZBError, NZBMetaEditor, NZBParser

nzbs = Path("tests/__nzbs__/")

invalid_xml = """\
<?xml version="1.0" encoding="iso-8859-1" ?>
<!DOCTYPE nzb PUBLIC "-//newzBin//DTD NZB 1.1//EN" "http://www.newzbin.com/DTD/nzb/nzb-1.1.dtd">
<nzb xmlns="http://www.newzbin.com/DTD/2003/nzb">
    <head>
        <meta type="title">Your File!</meta>
        <meta type="password">secret</meta>
        <meta type="tag">HD</meta>
        <meta type="category">TV</meta>
    </head>
    <file poster="Joe Bloggs &lt;bloggs@nowhere.example&gt;" date="1071674882" subject="Here's your file!  abc-mr2a.r01 (1/2)">
        <groups>
            <group>alt.binaries.newzbin</group>
            <group>alt.binaries.mojo</group>
        </groups>
        <segments>
            <segment bytes="102394" number="1">123456789abcdef@news.newzbin.com</segment>
            <segment bytes="4501" number="2">987654321fedbca@news.newzbin.com</segment>
        </segments>
    </file>
"""

valid_xml_but_invalid_nzb = """\
<?xml version="1.0" encoding="iso-8859-1" ?>
<!DOCTYPE nzb PUBLIC "-//newzBin//DTD NZB 1.1//EN" "http://www.newzbin.com/DTD/nzb/nzb-1.1.dtd">
<nzb xmlns="http://www.newzbin.com/DTD/2003/nzb">
    <head>
        <meta type="title">Your File!</meta>
    </head>
    <file poster="Joe Bloggs &lt;bloggs@nowhere.example&gt;" date="1071674882" subject="Here's your file!  abc-mr2a.r01 (1/2)">
        <groups>
            <group>alt.binaries.newzbin</group>
            <group>alt.binaries.mojo</group>
        </groups>
    </file>
</nzb>"""


def test_invalid_nzb_error() -> None:
    try:
        message = "Missing something in the NZB"
        raise InvalidNZBError(message)
    except InvalidNZBError as error:
        assert error.message == message
        assert str(error) == message
        assert repr(error) == 'InvalidNZBError("Missing something in the NZB")'


def test_saving_without_filename() -> None:
    with pytest.raises(FileNotFoundError):
        NZBMetaEditor((nzbs / "spec_example.nzb").read_text()).save()


def test_saving_overwrite() -> None:
    with pytest.raises(FileExistsError):
        NZBMetaEditor.from_file(nzbs / "spec_example.nzb").save()


def test_parsing_invalid_nzb() -> None:
    with pytest.raises(InvalidNZBError):
        NZBParser(invalid_xml).parse()

    with pytest.raises(InvalidNZBError):
        NZBParser(valid_xml_but_invalid_nzb).parse()


def test_editing_invalid_nzb() -> None:
    with pytest.raises(InvalidNZBError):
        NZBMetaEditor(invalid_xml)


def test_parser_exceptions() -> None:
    with pytest.raises(InvalidNZBError):
        NZBParser.from_file(nzbs / "malformed_files.nzb").parse()

    with pytest.raises(InvalidNZBError):
        NZBParser.from_file(nzbs / "malformed_files2.nzb").parse()

    with pytest.raises(InvalidNZBError):
        NZBParser.from_file(nzbs / "malformed_groups.nzb").parse()

    with pytest.raises(InvalidNZBError):
        NZBParser.from_file(nzbs / "malformed_segments.nzb").parse()



from __future__ import annotations

from importlib import metadata
from typing import NamedTuple


class Version(NamedTuple):
    major: int
    minor: int
    micro: int


__version__ = metadata.version("nzb")
__version_tuple__ = Version(*[int(i) for i in __version__.split(".")])



from __future__ import annotations

from nzb._utils import meta_constructor


def test_meta_constructor() -> None:
    assert meta_constructor(title="hello") == [{"@type": "title", "#text": "hello"}]
    assert meta_constructor(passwords="secret") == [{"@type": "password", "#text": "secret"}]
    assert meta_constructor(passwords=("secret", "secret2")) == [
        {"@type": "password", "#text": "secret"},
        {"@type": "password", "#text": "secret2"},
    ]
    assert meta_constructor(tags="hd") == [{"@type": "tag", "#text": "hd"}]
    assert meta_constructor(tags=("hd", "tv")) == [
        {"@type": "tag", "#text": "hd"},
        {"@type": "tag", "#text": "tv"},
    ]
    assert meta_constructor(category="cat") == [{"@type": "category", "#text": "cat"}]



from __future__ import annotations


class NZBException(Exception):
    """Base exception for all nzb exceptions"""


class InvalidNZBError(NZBException):
    """Invalid NZB"""

    def __init__(self, message: str) -> None:
        self.message = message
        """Human readable error message"""
        super().__init__(message)

    def __str__(self) -> str:
        """Equivalent to accessing the .member attribute"""
        return self.message

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}("{self.message}")'



from __future__ import annotations

import datetime
from pathlib import Path

import pytest

from nzb import File, NZBParser, Segment

nzbs = Path("tests/__nzbs__").resolve()


def test_spec_example_nzb() -> None:
    nzb = NZBParser.from_file(nzbs / "spec_example.nzb").parse()
    assert nzb.meta.title == "Your File!"
    assert nzb.meta.passwords == ("secret",)
    assert nzb.meta.tags == ("HD",)
    assert nzb.meta.password == "secret"
    assert nzb.meta.tag == "HD"
    assert nzb.meta.category == "TV"
    assert len(nzb.files) == 1
    assert nzb.is_rar() is True
    assert nzb.is_obfuscated() is True
    assert nzb.file.name == "abc-mr2a.r01"
    assert nzb.file.stem == "abc-mr2a"
    assert nzb.file.suffix == ".r01"
    assert nzb.size == 106895
    assert len(nzb.files[0].segments) == 2
    assert set(nzb.files[0].segments) == set(
        (
            Segment(size=102394, number=1, message_id="123456789abcdef@news.newzbin.com"),
            Segment(size=4501, number=2, message_id="987654321fedbca@news.newzbin.com"),
        )
    )
    assert set(nzb.files[0].groups) == set(("alt.binaries.mojo", "alt.binaries.newzbin"))


def test_big_buck_bunny() -> None:
    nzb = NZBParser.from_file(nzbs / "big_buck_bunny.nzb").parse()

    assert nzb.meta.title is None
    assert nzb.meta.passwords is None
    assert nzb.meta.tags is None
    assert nzb.meta.password is None
    assert nzb.meta.tag is None
    assert nzb.meta.category is None
    assert len(nzb.files) == 5
    assert nzb.is_rar() is False
    assert nzb.is_obfuscated() is False
    assert nzb.has_par2() is True
    assert nzb.file.name == "Big Buck Bunny - S01E01.mkv"
    assert nzb.file.stem == "Big Buck Bunny - S01E01"
    assert nzb.file.suffix == ".mkv"
    assert nzb.size == 22704889
    assert set(nzb.names) == {
        "Big Buck Bunny - S01E01.mkv.vol03+04.par2",
        "Big Buck Bunny - S01E01.mkv.vol01+02.par2",
        "Big Buck Bunny - S01E01.mkv.par2",
        "Big Buck Bunny - S01E01.mkv",
        "Big Buck Bunny - S01E01.mkv.vol00+01.par2",
    }
    assert set(nzb.stems) == {
        "Big Buck Bunny - S01E01",
        "Big Buck Bunny - S01E01.mkv",
        "Big Buck Bunny - S01E01.mkv.vol01+02",
        "Big Buck Bunny - S01E01.mkv.vol03+04",
        "Big Buck Bunny - S01E01.mkv.vol00+01",
    }
    assert set(nzb.suffixes) == {".mkv", ".par2"}
    assert set(nzb.posters) == {"John <nzb@nowhere.example>"}
    assert set(nzb.groups) == {"alt.binaries.boneless"}
    assert nzb.par2_size == 5183128
    assert nzb.par2_percentage == pytest.approx(22, 1.0)
    assert nzb.file == File(
        poster="John <nzb@nowhere.example>",
        datetime=datetime.datetime(2024, 1, 28, 11, 18, 28, tzinfo=datetime.timezone.utc),
        subject='[1/5] - "Big Buck Bunny - S01E01.mkv" yEnc (1/24) 16981056',
        groups=("alt.binaries.boneless",),
        segments=(
            Segment(size=739067, number=1, message_id="9cacde4c986547369becbf97003fb2c5-9483514693959@example"),
            Segment(size=739549, number=2, message_id="70a3a038ce324e618e2751e063d6a036-7285710986748@example"),
            Segment(size=739728, number=3, message_id="a209875cefd44440aa91590508b48f5b-4625756912881@example"),
            Segment(size=739664, number=4, message_id="44057720ed4e45e4bce21d53249d03f8-8250738040266@example"),
            Segment(size=739645, number=5, message_id="cfc13d14583c484483aa49ac420bad27-9491395432062@example"),
            Segment(size=739538, number=6, message_id="5e90857531be401e9d0b632221fe2fb7-9854527985639@example"),
            Segment(size=739708, number=7, message_id="c33a2bba79494840a09d750b19d3b287-2550637855678@example"),
            Segment(size=739490, number=8, message_id="38006019d94f4ecc8f19c389c00f1ebe-7841585708380@example"),
            Segment(size=739667, number=9, message_id="b75a2425bef24fd5affb00dc3db789f6-7051027232703@example"),
            Segment(size=739540, number=10, message_id="79a027e3bfde458ea2bd0db1632fc84e-7270120407913@example"),
            Segment(size=739657, number=11, message_id="fb2bd74e1257487a9240ef0cf81765cc-7147741101314@example"),
            Segment(size=739647, number=12, message_id="d39ca8be78c34e3fa6f3211f1b397b3a-4725950858191@example"),
            Segment(size=739668, number=13, message_id="a4c15599055848dda1eff3b6b406fa78-8111735210252@example"),
            Segment(size=739721, number=14, message_id="2f1cec363ed24584b4127af86ac312ad-7204153818612@example"),
            Segment(size=739740, number=15, message_id="30ff3514896543a8ac91ec80346a5d40-9134304686352@example"),
            Segment(size=739538, number=16, message_id="1f75cfa20d884b5b972cfd2e9ebef249-8919850122587@example"),
            Segment(size=739646, number=17, message_id="8e22b0f973de4393a0a30ab094565316-6722799721412@example"),
            Segment(size=739610, number=18, message_id="faddf83650cc4de1a8bee68cffca40a1-5979589815618@example"),
            Segment(size=739514, number=19, message_id="6b8c23e43d4240da812b547babdc0423-6409257710918@example"),
            Segment(size=739920, number=20, message_id="802bd0dcef134ac690044e0a09fece60-8492061912475@example"),
            Segment(size=739634, number=21, message_id="efc4b3966a1f4b7787677e9e9a214727-5444471572012@example"),
            Segment(size=739691, number=22, message_id="247efca709114fd181bcaef0f487925f-4076317880026@example"),
            Segment(size=739638, number=23, message_id="665d9fc5edba4faca68ae835b702b4c7-9814601723860@example"),
            Segment(size=510541, number=24, message_id="962fddf3e07444988731b52aeaa9b2aa-1283919353788@example"),
        ),
    )


def test_valid_nzb_with_one_missing_segment() -> None:
    nzb = NZBParser.from_file(nzbs / "valid_nzb_with_one_missing_segment.nzb").parse()

    assert nzb.file == File(
        poster="John <nzb@nowhere.example>",
        datetime=datetime.datetime(2024, 1, 28, 11, 18, 28, tzinfo=datetime.timezone.utc),
        subject='[1/5] - "Big Buck Bunny - S01E01.mkv" yEnc (1/24) 16981056',
        groups=("alt.binaries.boneless",),
        segments=(
            Segment(size=739067, number=1, message_id="9cacde4c986547369becbf97003fb2c5-9483514693959@example"),
            Segment(size=739549, number=2, message_id="70a3a038ce324e618e2751e063d6a036-7285710986748@example"),
            Segment(size=739728, number=3, message_id="a209875cefd44440aa91590508b48f5b-4625756912881@example"),
            Segment(size=739664, number=4, message_id="44057720ed4e45e4bce21d53249d03f8-8250738040266@example"),
            Segment(size=739645, number=5, message_id="cfc13d14583c484483aa49ac420bad27-9491395432062@example"),
            Segment(size=739538, number=6, message_id="5e90857531be401e9d0b632221fe2fb7-9854527985639@example"),
            Segment(size=739708, number=7, message_id="c33a2bba79494840a09d750b19d3b287-2550637855678@example"),
            Segment(size=739490, number=8, message_id="38006019d94f4ecc8f19c389c00f1ebe-7841585708380@example"),
            Segment(size=739667, number=9, message_id="b75a2425bef24fd5affb00dc3db789f6-7051027232703@example"),
            Segment(size=739540, number=10, message_id="79a027e3bfde458ea2bd0db1632fc84e-7270120407913@example"),
            Segment(size=739657, number=11, message_id="fb2bd74e1257487a9240ef0cf81765cc-7147741101314@example"),
            Segment(size=739647, number=12, message_id="d39ca8be78c34e3fa6f3211f1b397b3a-4725950858191@example"),
            # 13th Segment is missing here
            Segment(size=739721, number=14, message_id="2f1cec363ed24584b4127af86ac312ad-7204153818612@example"),
            Segment(size=739740, number=15, message_id="30ff3514896543a8ac91ec80346a5d40-9134304686352@example"),
            Segment(size=739538, number=16, message_id="1f75cfa20d884b5b972cfd2e9ebef249-8919850122587@example"),
            Segment(size=739646, number=17, message_id="8e22b0f973de4393a0a30ab094565316-6722799721412@example"),
            Segment(size=739610, number=18, message_id="faddf83650cc4de1a8bee68cffca40a1-5979589815618@example"),
            Segment(size=739514, number=19, message_id="6b8c23e43d4240da812b547babdc0423-6409257710918@example"),
            Segment(size=739920, number=20, message_id="802bd0dcef134ac690044e0a09fece60-8492061912475@example"),
            Segment(size=739634, number=21, message_id="efc4b3966a1f4b7787677e9e9a214727-5444471572012@example"),
            Segment(size=739691, number=22, message_id="247efca709114fd181bcaef0f487925f-4076317880026@example"),
            Segment(size=739638, number=23, message_id="665d9fc5edba4faca68ae835b702b4c7-9814601723860@example"),
            Segment(size=510541, number=24, message_id="962fddf3e07444988731b52aeaa9b2aa-1283919353788@example"),
        ),
    )


def test_bad_subject() -> None:
    nzb = NZBParser.from_file(nzbs / "bad_subject.nzb").parse()
    assert nzb.files[0].name == ""
    assert nzb.files[0].stem == ""
    assert nzb.files[0].suffix == ""
    assert nzb.files[0].is_par2() is False
    assert nzb.files[0].is_rar() is False
    assert nzb.is_rar() is False
    assert nzb.has_par2() is False
    assert nzb.is_obfuscated() is True


def test_non_standard_meta() -> None:
    nzb = NZBParser.from_file(nzbs / "non_standard_meta.nzb").parse()
    assert nzb.meta.title is None
    assert nzb.meta.passwords is None
    assert nzb.meta.tags is None
    assert nzb.meta.category is None


def test_single_rar_nzb() -> None:
    nzb = NZBParser.from_file(nzbs / "one_rar_file.nzb").parse()
    assert nzb.has_rar() is True
    assert nzb.is_rar() is False
    assert nzb.has_par2() is False


def test_multi_rar_nzb() -> None:
    nzb = NZBParser.from_file(nzbs / "multi_rar.nzb").parse()
    assert nzb.has_rar() is True
    assert nzb.is_rar() is True
    assert nzb.has_par2() is False



"""
https://sabnzbd.org/wiki/extra/nzb-spec
https://web.archive.org/web/20240709113825/https://sabnzbd.org/wiki/extra/nzb-spec
"""

from __future__ import annotations

import re
from typing import Any, TypeAlias, Union, cast

from natsort import natsorted

from nzb._exceptions import InvalidNZBError
from nzb._models import File, Meta, Segment


def parse_metadata(nzb: dict[str, Any]) -> Meta:
    """
    Parses the <meta>...</meta> field present in an NZB.

    ```xml
    <?xml version="1.0" encoding="iso-8859-1" ?>
    <!DOCTYPE nzb PUBLIC "-//newzBin//DTD NZB 1.1//EN" "http://www.newzbin.com/DTD/nzb/nzb-1.1.dtd">
    <nzb xmlns="http://www.newzbin.com/DTD/2003/nzb">
        <head>
            <meta type="title">Your File!</meta>
            <meta type="password">secret</meta>
            <meta type="tag">HD</meta>
            <meta type="category">TV</meta>
        </head>
    </nzb>
    ```
    """

    meta = nzb.get("nzb", {}).get("head", {}).get("meta")
    # There's 3 possible things that we can get from the above here:
    # - A list of dictionaries if there's more than 1 meta field present
    # - A dictionary if there's only one meta field present
    # - None if there's no meta field

    # Here's the type representation of the three possible cases that we need to handle
    MetaFieldType: TypeAlias = Union[list[dict[str, str]], dict[str, str], None]

    # Explicit cast to tell typecheckers the return type based on the above 3 points.
    meta = cast(MetaFieldType, meta)

    if meta is None:
        # Meta is optional, so we will not error
        # just return an instance with all values set to None
        return Meta()

    if isinstance(meta, dict):
        meta = [meta]

    passwordset = set()
    tagset = set()
    title = None
    name = None
    category = None

    for item in meta:
        if item.get("@type", "").casefold() == "title":
            title = item.get("#text")

        if item.get("@type", "").casefold() == "password":
            # spec allows for multiple passwords by repeating the same field
            # <meta type="password">secret1</meta>
            # <meta type="password">secret2</meta>
            # <meta type="password">secret3</meta>
            if password := item.get("#text"):
                passwordset.add(password)

        if item.get("@type", "").casefold() == "tag":
            # spec allows for multiple tags by repeating the same field
            # <meta type="tag">HD</meta>
            # <meta type="tag">Anime</meta>
            # <meta type="tag">1080p</meta>
            if tag := item.get("#text"):
                tagset.add(tag.strip())

        if item.get("@type", "").casefold() == "category":
            category = item.get("#text")

    return Meta(
        title=title or name,
        passwords=passwordset if passwordset else None,  # type: ignore
        tags=tagset if tagset else None,  # type: ignore
        category=category,
    )


def parse_segments(segmentdict: dict[str, list[dict[str, str]] | dict[str, str] | None] | None) -> tuple[Segment, ...]:
    """
    Parses the <segments>...</segments> field present in an NZB.

    There's 3 possible things that we can get here:
    - A list of dictionaries if there's more than 1 segment field present
    - A dictionary if there's only one segment field present
    - None if there's no segment field

    ```xml
    <?xml version="1.0" encoding="iso-8859-1" ?>
    <!DOCTYPE nzb PUBLIC "-//newzBin//DTD NZB 1.1//EN" "http://www.newzbin.com/DTD/nzb/nzb-1.1.dtd">
    <nzb xmlns="http://www.newzbin.com/DTD/2003/nzb">
        <file poster="Joe Bloggs &lt;bloggs@nowhere.example&gt;" date="1071674882" subject="Here's your file!  abc-mr2a.r01 (1/2)">
            <groups>[...]</groups>
            <segments>
                <segment bytes="102394" number="1">123456789abcdef@news.newzbin.com</segment>
                <segment bytes="4501" number="2">987654321fedbca@news.newzbin.com</segment>
            </segments>
        </file>
    </nzb>
    ```
    """
    # It's nested or possibly None
    segments = segmentdict.get("segment") if segmentdict else None

    if segments is None:
        raise InvalidNZBError("Missing or malformed <segments>...</segments>!")

    if isinstance(segments, dict):
        segments = [segments]

    segmentset: set[Segment] = set()

    for segment in segments:
        try:
            size = segment["@bytes"]
            number = segment["@number"]
            message_id = segment["#text"]
        except KeyError:
            # This segment is broken
            # We do not error here because a few missing
            # segments don't invalidate the nzb.
            continue

        segmentset.add(Segment(size=size, number=number, message_id=message_id))  # type: ignore

    return tuple(natsorted(segmentset, key=lambda seg: seg.number))


def parse_files(nzb: dict[str, Any]) -> tuple[File, ...]:
    """
    Parses the <file>...</file> field present in an NZB.

    ```xml
    <?xml version="1.0" encoding="iso-8859-1" ?>
    <!DOCTYPE nzb PUBLIC "-//newzBin//DTD NZB 1.1//EN" "http://www.newzbin.com/DTD/nzb/nzb-1.1.dtd">
    <nzb xmlns="http://www.newzbin.com/DTD/2003/nzb">
        <file poster="Joe Bloggs &lt;bloggs@nowhere.example&gt;" date="1071674882" subject="Here's your file!  abc-mr2a.r01 (1/2)">
            <groups>[...]</groups>
            <segments>[...]</segments>
        </file>
    </nzb>
    ```
    """

    files = nzb.get("nzb", {}).get("file")
    # There's 3 possible things that we can get from the above here:
    # - A list of dictionaries if there's more than 1 file field present, i.e, list[dict[str, str]]
    # - A dictionary if there's only one file field present, i.e, dict[str, str]
    # - None if there's no meta field

    # Here's the type representation of the three possible cases that we need to handle
    FileFieldType: TypeAlias = Union[list[dict[str, str]], dict[str, str], None]

    # Explicit cast to tell typecheckers the return type based on the above 3 points.
    files = cast(FileFieldType, files)

    if files is None:
        raise InvalidNZBError("Missing or malformed <file>...</file>!")

    if isinstance(files, dict):
        files = [files]

    fileset: set[File] = set()

    for file in files:
        groupset: set[str] = set()

        groups = file.get("groups").get("group") if file.get("groups") else None
        # There's 3 possible things that we can get from the above here:
        # - A list of strings if there's more than 1 group present, i.e, list[str]
        # - A string if there's only one file field present, i.e, str
        # - None if there's no group field

        # Here's the type representation of the three possible cases that we need to handle
        GroupFieldType: TypeAlias = Union[list[str], str, None]

        # Explicit cast to tell typecheckers the return type based on the above 3 points.
        groups = cast(GroupFieldType, groups)

        if groups is None:
            raise InvalidNZBError("Missing or malformed <groups>...</groups>!")

        if isinstance(groups, str):
            groupset.add(groups)
        else:
            groupset.update(groups)

        fileset.add(
            File(
                poster=file.get("@poster"),
                datetime=file.get("@date"),
                subject=file.get("@subject"),
                groups=natsorted(groupset),  # type: ignore
                segments=parse_segments(file.get("segments")),
            )
        )

    return tuple(natsorted(fileset, key=lambda file: file.subject))


def parse_doctype(nzb: str) -> str | None:
    """
    Parses the DOCTYPE from an NZB file.

    Quoting https://www.oreilly.com/library/view/xml-pocket-reference/9780596100506/ch01s02s09.html:
    "A document type or DOCTYPE declaration provides information to a validating XML parser about how to validate an XML document.
    The DOCTYPE keyword appears first; then the document, or root, element of the document being validated is identified,
    followed by either a SYSTEM or PUBLIC identifier."

    Example:
    ```xml
    <?xml version="1.0" encoding="UTF-8"?>
    <!DOCTYPE nzb PUBLIC "-//newzBin//DTD NZB 1.1//EN" "http://www.newzbin.com/DTD/nzb/nzb-1.1.dtd">
    <nzb xmlns="http://www.newzbin.com/DTD/2003/nzb">
    ...
    </nzb>
    ```
    """
    doctype = re.search(r"<!DOCTYPE nzb.*>", nzb, re.IGNORECASE)
    return doctype.group() if doctype else None



from __future__ import annotations

import re
from functools import cached_property
from os.path import splitext

from natsort import natsorted
from pydantic import BaseModel, ByteSize, ConfigDict

from nzb._types import UTCDateTime
from nzb._utils import (
    name_is_par2,
    name_is_rar,
    stem_is_obfuscated,
)


class ParentModel(BaseModel):
    model_config = ConfigDict(frozen=True, str_strip_whitespace=True)


class Meta(ParentModel):
    """Optional creator-definable metadata for the contents of the NZB."""

    title: str | None = None
    """Title."""

    passwords: tuple[str, ...] | None = None  # Can be specified multiple times.
    """Password(s)."""

    tags: tuple[str, ...] | None = None  # Can be specified multiple times.
    """Tag(s)."""

    category: str | None = None
    """Category."""

    @cached_property
    def password(self) -> str | None:
        """
        Return the first password from [`Meta.passwords`][nzb._models.Meta.passwords]
        if it exists, None otherwise.

        This is essentially just syntactic sugar for `password = passwords[0] if passwords else None`
        because although the spec allows multiple passwords, single passwords are far more common.
        """
        return self.passwords[0] if self.passwords else None

    @cached_property
    def tag(self) -> str | None:
        """
        The first tag from [`Meta.tags`][nzb._models.Meta.tags]
        if it exists, None otherwise.

        This is essentially just syntactic sugar for `tag = tags[0] if tags else None`
        because although the spec allows multiple tags, single tags are far more common.
        """
        return self.tags[0] if self.tags else None


class Segment(ParentModel):
    """One part segment of a file."""

    size: ByteSize
    """Size of the segment."""
    number: int
    """Number of the segment."""
    message_id: str
    """Message ID of the segment."""


class File(ParentModel):
    """Represents a complete file, consisting of segments that make up a file."""

    poster: str
    """The poster of the file."""

    datetime: UTCDateTime
    """The date and time when the file was posted, in UTC."""

    subject: str
    """The subject of the file."""  # Ideally it contains the filename, segment count, and other relevant information.

    groups: tuple[str, ...]  # Every file must have atleast one group.
    """Groups that reference the file."""

    segments: tuple[Segment, ...]  # Every file must have atleast one segment.
    """Segments that make up the file."""

    @cached_property
    def size(self) -> ByteSize:
        """Size of the file calculated from the sum of segment sizes."""
        return ByteSize(sum(segment.size for segment in self.segments))

    @cached_property
    def name(self) -> str:
        """
        Complete name of the file with it's extension extracted from the subject.
        May return an empty string if it fails to extract the name.
        """
        # https://github.com/sabnzbd/sabnzbd/blob/02b4a116dd4b46b2d2f33f7bbf249f2294458f2e/sabnzbd/nzbstuff.py#L104-L106
        if parsed := re.search(r'"([^"]*)"', self.subject):
            return parsed.group(1).strip()
        elif parsed := re.search(
            r"\b([\w\-+()' .,]+(?:\[[\w\-/+()' .,]*][\w\-+()' .,]*)*\.[A-Za-z0-9]{2,4})\b", self.subject
        ):
            return parsed.group(1).strip()
        else:
            return ""

    @cached_property
    def stem(self) -> str:
        """
        Base name of the file without it's extension extracted from the [`File.name`][nzb._models.File.name].
        May return an empty string if it fails to extract the stem.
        """
        if not self.name:
            return ""
        else:
            root, _ = splitext(self.name)
            return root

    @cached_property
    def suffix(self) -> str:
        """
        Extension of the file extracted from the [`File.name`][nzb._models.File.name].
        May return an empty string if it fails to extract the extension.
        """
        if not self.name:
            return ""
        else:
            _, ext = splitext(self.name)
            return ext

    def is_par2(self) -> bool:
        """
        Return `True` if the file is a `.par2` file, `False` otherwise.
        """
        return name_is_par2(self.name)

    def is_rar(self) -> bool:
        """
        Return `True` if the file is a `.rar` file, `False` otherwise.
        """
        return name_is_rar(self.name)

    def is_obfuscated(self) -> bool:
        """
        Return `True` if the file is obfuscated, `False` otherwise.
        """
        return stem_is_obfuscated(self.stem)


class NZB(ParentModel):
    """Represents a complete NZB file."""

    meta: Meta = Meta()
    """Optional creator-definable metadata for the contents of the NZB."""

    files: tuple[File, ...]
    """File objects representing the files included in the NZB."""

    @cached_property
    def file(self) -> File:
        """
        The main content file (episode, movie, etc) in the NZB.
        This is determined by finding the largest file in the NZB
        and may not always be accurate.
        """
        return max(self.files, key=lambda file: file.size)

    @cached_property
    def size(self) -> ByteSize:
        """Total size of all the files in the NZB."""
        return ByteSize(sum(file.size for file in self.files))

    @cached_property
    def names(self) -> tuple[str, ...]:
        """
        Tuple of unique file names across all the files in the NZB.
        May return an empty tuple if it fails to extract the name for every file.
        """
        return tuple(natsorted({file.name for file in self.files}))

    @cached_property
    def stems(self) -> tuple[str, ...]:
        """
        Tuple of unique file stems (basename) across all the files in the NZB.
        May return an empty tuple if it fails to extract the stem for every file.
        """
        return tuple(natsorted({file.stem for file in self.files}))

    @cached_property
    def suffixes(self) -> tuple[str, ...]:
        """
        Tuple of unique file extensions across all the files in the NZB.
        May return an empty tuple if it fails to extract the extension for every file.
        """
        return tuple(natsorted({file.suffix for file in self.files}))

    @cached_property
    def posters(self) -> tuple[str, ...]:
        """
        Tuple of unique posters across all the files in the NZB.
        """
        return tuple(natsorted({file.poster for file in self.files}))

    @cached_property
    def groups(self) -> tuple[str, ...]:
        """
        Tuple of unique groups across all the files in the NZB.
        """
        groupset: set[str] = set()

        for file in self.files:
            groupset.update(file.groups)

        return tuple(natsorted(groupset))

    @cached_property
    def par2_size(self) -> ByteSize:
        """
        Total size of all the `.par2` files.
        """
        return ByteSize(sum(file.size for file in self.files if file.is_par2()))

    @cached_property
    def par2_percentage(self) -> float:
        """
        Percentage of the size of all the `.par2` files relative to the total size.
        """
        return (self.par2_size / self.size) * 100

    def has_rar(self) -> bool:
        """
        Return `True` if any file in the NZB is a `.rar` file, `False` otherwise.
        """
        return any(file.is_rar() for file in self.files)

    def is_rar(self) -> bool:
        """
        Return `True` if all files in the NZB are `.rar` files, `False` otherwise.
        """
        return all(file.is_rar() for file in self.files)

    def is_obfuscated(self) -> bool:
        """
        Return `True` if any file in the NZB is obfuscated, `False` otherwise.
        """
        return any(file.is_obfuscated() for file in self.files)

    def has_par2(self) -> bool:
        """
        Return `True` if there's at least one `.par2` file in the NZB, `False` otherwise.
        """
        return any(file.is_par2() for file in self.files)



from __future__ import annotations

from nzb._core import NZBMetaEditor, NZBParser
from nzb._exceptions import InvalidNZBError, NZBException
from nzb._models import NZB, File, Meta, Segment
from nzb._version import __version__, __version_tuple__

__all__ = (
    "NZBParser",
    "NZBMetaEditor",
    "Meta",
    "File",
    "Segment",
    "NZB",
    "NZBException",
    "InvalidNZBError",
    "__version__",
    "__version_tuple__",
)
